from zoo.llm.common.performance.utils import TiledGEMM, TiledMatrix
import math
from archx.utils import get_prod
from collections import OrderedDict
import inspect

def offchip_gemm_scheduling(batch, m, k, n, architecture_dict: OrderedDict, workload_dict: OrderedDict) -> TiledGEMM:
    """
    Sets tiling for dram <-> sram for single-node and multi-node configurations
    """
    # Retrieve dicts
    arch = workload_dict['architecture']
    isram_dict = architecture_dict['isram']
    wsram_dict = architecture_dict['wsram']
    osram_dict = architecture_dict['osram']
    input_bitwidth = workload_dict['activation_bitwidth']
    weight_bitwidth = workload_dict['weight_bitwidth']
    output_bitwidth = workload_dict['activation_bitwidth']
    stationary = workload_dict['noc_stationary']
    array_width = architecture_dict['ififo']['instance'][-1]
    array_height = architecture_dict['wfifo']['instance'][-1]

    # SRAM configurations
    # divide banks by to buffer
    isram_bank = isram_dict['query']['bank'] / 2
    wsram_bank = wsram_dict['query']['bank'] / 2
    osram_bank = osram_dict['query']['bank'] / 2

    isram_width = isram_dict['query']['width']
    wsram_width = wsram_dict['query']['width']
    osram_width = osram_dict['query']['width']

    isram_depth = isram_dict['query']['depth']
    wsram_depth = wsram_dict['query']['depth']
    osram_depth = osram_dict['query']['depth']

    isram_size = isram_bank * isram_width * isram_depth
    wsram_size = wsram_bank * wsram_width * wsram_depth
    osram_size = osram_bank * osram_width * osram_depth

    isram_elements = isram_size / input_bitwidth
    wsram_elements = wsram_size / weight_bitwidth
    osram_elements = osram_size / output_bitwidth

    # number of nodes
    nodes = isram_dict['instance'][0] * isram_dict['instance'][1] if 'irouter' in architecture_dict else 1

    if stationary == 'os':
        # Output stationary scheduling
        # initialize m tile size to m (in llms, very oftem batch size or smaller)
        # initialize n tile size to maximum size that fully utalizes noc
        tile_m = min(m, array_width)
        tile_n = ((n * batch) / nodes)

        # if tile_n does not fit in osram, reduce tile_n
        while(tile_n * tile_m > osram_elements):
            tile_n /= 2

        # if small tile n, and is smaller than array, increase tile_n
        while((tile_n * tile_m * 2 < osram_elements) and (tile_n < array_height)):
            tile_n *= 2

        # if small tile n, and can increase m, increase tile_m
        # multiply by 2 to check if you can increase tile_m
        while((tile_m * tile_n * 2 < osram_elements) and (tile_m < m)):
            tile_m *= 2

        # initialize k tile size to maximum size that fully utalizes wsram
        tile_k = wsram_elements / tile_n

        # if tile_k does not fit in isram, reduce tile_k
        while(tile_k * tile_m > isram_elements):
            tile_k /= 2

        while arch == 'tensor' and (tile_k < array_width):
              tile_k *= 2
              tile_n /= 2

    elif stationary == 'ws':
        # Weight stationary scheduling
        # initialize m tile size to m (in llms, very oftem batch size or smaller)
        # initialize k tile size to maximum size that fully utalizes noc
        tile_m = min(m, array_width)
        tile_k = ((k * batch) / nodes)

        # if tile_k does not fit in isram, reduce tile_k
        while(tile_k * tile_m > isram_elements):
            tile_k /= 2

        # initialize n tile size to maximum size that fully utalizes wsram
        tile_n = wsram_elements / tile_k

        # if tile_n does not fit in osram or wsram, reduce tile_n
        while(tile_n > osram_elements / tile_m):
            tile_n /= 2

        # if small tile k, and can increase m, increase tile_m
        # multiply by 2 to check if you can increase tile_m
        while((tile_m * tile_k * 2 < isram_elements) and (tile_m < m)):
            tile_m *= 2

    elif stationary == 'is':
        # Input stationary scheduling
        # initialize k tile size to m(quantized dim, smaller data size)
        # initialize m tile size to maximum size that fully utalizes noc
        tile_k = min(k, array_height)
        tile_m = ((m * batch) / nodes)

        # if tile_m does not fit in isram, reduce tile_m
        while(tile_m * tile_k > isram_elements):
            tile_m /= 2

        # initialize n tile size to maximum size that fully utalizes osram
        tile_n = osram_elements / tile_m
        
        # if tile_n does not fit in wsram, reduce tile_n
        while(tile_n * tile_k > wsram_elements):
            tile_n /= 2

        # if small tile k, and can increase k, increase tile_k
        # multiply by 2 to check if you can increase tile_k
        while((tile_k * tile_m * 2 < isram_elements) and (tile_k < k)):
            tile_k *= 2

    if tile_m * tile_k > isram_elements:
        raise ValueError('Tile size exceeds isram size')
    if tile_k * tile_n > wsram_elements:
        raise ValueError('Tile size exceeds wsram size')
    if tile_m * tile_n > osram_elements:
        raise ValueError('Tile size exceeds osram size')

    tile_m = 2 ** math.ceil(math.log2(tile_m))
    tile_k = 2 ** math.ceil(math.log2(tile_k))
    tile_n = 2 ** math.ceil(math.log2(tile_n))

    tile_m = int(tile_m)
    tile_k = int(tile_k)
    tile_n = int(tile_n)

    tiles = TiledGEMM(batch=batch, m=m, k=k, n=n, tile_m=tile_m, tile_k=tile_k, tile_n=tile_n, m_k_bitwidth=input_bitwidth, k_n_bitwidth=weight_bitwidth, m_n_bitwidth=output_bitwidth)
    if not tiles.is_valid:
        raise ValueError('Invalid tiling configuration')

    return tiles

def offchip_gemm_events(tiles: TiledGEMM, architecture_dict: OrderedDict, workload_dict: OrderedDict) -> OrderedDict:
    """
    Estimates dram and sram reads and writes to from offchip memory to onchip memory for GEMM operations
    """
    # Retrieve dicts
    stationary = workload_dict['noc_stationary']
    isram_width = architecture_dict['isram']['query']['width']
    wsram_width = architecture_dict['wsram']['query']['width']
    osram_width = architecture_dict['osram']['query']['width']

    if stationary == 'os':
        # output stationary scheduling
        # M x K Input matrix dram <-> isram events (dram reads, isram writes)
        # input = batch x m_tiles x k_tiles x n_tiles x ceil(tile_bits / isram_width) -> (applies to all, but with flow adjusted for each, this is an output stationary example)
        # weight = batch x k_tiles x n_tiles x m_tiles x ceil(tile_bits / wsram_width)
        # output = batch x m_tiles x n_tiles x ceil(tile_bits / osram_width) -> (no k dim, as it's output stationary. Output stationary maps noc to outputs, so no need to map to k)
        # More detailed breakdown, allows for instances where the tile size is smaller than sram width, which increases events compared to using total_bits/width.
        m_full_k_full_events = tiles.m_full_k_full_total_tiles * tiles.n_tiles * math.ceil(tiles.m_full_k_full_tile_bits / isram_width)
        m_full_k_partial_events = tiles.m_full_k_partial_total_tiles * tiles.n_tiles * math.ceil(tiles.m_full_k_partial_tile_bits / isram_width)
        m_partial_k_full_events = tiles.m_partial_k_full_total_tiles * tiles.n_tiles * math.ceil(tiles.m_partial_k_full_tile_bits / isram_width)
        m_partial_k_partial_events = tiles.m_partial_k_partial_total_tiles * tiles.n_tiles * math.ceil(tiles.m_partial_k_partial_tile_bits / isram_width)

        # K x N Weight matrix dram <-> wsram events (dram reads, wsram writes)
        k_full_n_full_events = tiles.k_full_n_full_total_tiles * tiles.m_tiles * math.ceil(tiles.k_full_n_full_tile_bits / wsram_width)
        k_full_n_partial_events = tiles.k_full_n_partial_total_tiles * tiles.m_tiles * math.ceil(tiles.k_full_n_partial_tile_bits / wsram_width)
        k_partial_n_full_events = tiles.k_partial_n_full_total_tiles * tiles.m_tiles * math.ceil(tiles.k_partial_n_full_tile_bits / wsram_width)
        k_partial_n_partial_events = tiles.k_partial_n_partial_total_tiles * tiles.m_tiles * math.ceil(tiles.k_partial_n_partial_tile_bits / wsram_width)

        # M x N Output matrix osram <-> dram events (osram reads)
        m_full_n_full_events = tiles.m_full_n_full_total_tiles * math.ceil(tiles.m_full_n_full_tile_bits / osram_width)
        m_full_n_partial_events = tiles.m_full_n_partial_total_tiles * math.ceil(tiles.m_full_n_partial_tile_bits / osram_width)
        m_partial_n_full_events = tiles.m_partial_n_full_total_tiles * math.ceil(tiles.m_partial_n_full_tile_bits / osram_width)
        m_partial_n_partial_events = tiles.m_partial_n_partial_total_tiles * math.ceil(tiles.m_partial_n_partial_tile_bits / osram_width)

    elif stationary == 'is':
        # input stationary scheduling
        # M x K Input matrix dram <-> isram events (dram reads, isram writes)
        m_full_k_full_events = tiles.m_full_k_full_total_tiles * math.ceil(tiles.m_full_k_full_tile_bits / isram_width)
        m_full_k_partial_events = tiles.m_full_k_partial_total_tiles * math.ceil(tiles.m_full_k_partial_tile_bits / isram_width)
        m_partial_k_full_events = tiles.m_partial_k_full_total_tiles * math.ceil(tiles.m_partial_k_full_tile_bits / isram_width)
        m_partial_k_partial_events = tiles.m_partial_k_partial_total_tiles * math.ceil(tiles.m_partial_k_partial_tile_bits / isram_width)

        # K x N Weight matrix dram <-> wsram events (dram reads, wsram writes)
        k_full_n_full_events = tiles.k_full_n_full_total_tiles * tiles.m_tiles * math.ceil(tiles.k_full_n_full_tile_bits / wsram_width)
        k_full_n_partial_events = tiles.k_full_n_partial_total_tiles * tiles.m_tiles * math.ceil(tiles.k_full_n_partial_tile_bits / wsram_width)
        k_partial_n_full_events = tiles.k_partial_n_full_total_tiles * tiles.m_tiles * math.ceil(tiles.k_partial_n_full_tile_bits / wsram_width)
        k_partial_n_partial_events = tiles.k_partial_n_partial_total_tiles * tiles.m_tiles * math.ceil(tiles.k_partial_n_partial_tile_bits / wsram_width)

        # M x N Output matrix osram <-> dram events (osram reads)
        m_full_n_full_events = tiles.m_full_n_full_total_tiles * tiles.k_tiles * math.ceil(tiles.m_full_n_full_tile_bits / osram_width)
        m_full_n_partial_events = tiles.m_full_n_partial_total_tiles * tiles.k_tiles * math.ceil(tiles.m_full_n_partial_tile_bits / osram_width)
        m_partial_n_full_events = tiles.m_partial_n_full_total_tiles * tiles.k_tiles * math.ceil(tiles.m_partial_n_full_tile_bits / osram_width)
        m_partial_n_partial_events = tiles.m_partial_n_partial_total_tiles * tiles.k_tiles * math.ceil(tiles.m_partial_n_partial_tile_bits / osram_width)

    elif stationary == 'ws':
        # weight stationary scheduling
        # M x K Input matrix dram <-> isram events (dram reads, isram writes)
        m_full_k_full_events = tiles.m_full_k_full_total_tiles * tiles.n_tiles * math.ceil(tiles.m_full_k_full_tile_bits / isram_width)
        m_full_k_partial_events = tiles.m_full_k_partial_total_tiles * tiles.n_tiles * math.ceil(tiles.m_full_k_partial_tile_bits / isram_width)
        m_partial_k_full_events = tiles.m_partial_k_full_total_tiles * tiles.n_tiles * math.ceil(tiles.m_partial_k_full_tile_bits / isram_width)
        m_partial_k_partial_events = tiles.m_partial_k_partial_total_tiles * tiles.n_tiles * math.ceil(tiles.m_partial_k_partial_tile_bits / isram_width)

        # K x N Weight matrix dram <-> wsram events (dram reads, wsram writes)
        k_full_n_full_events = tiles.k_full_n_full_total_tiles * math.ceil(tiles.k_full_n_full_tile_bits / wsram_width)
        k_full_n_partial_events = tiles.k_full_n_partial_total_tiles * math.ceil(tiles.k_full_n_partial_tile_bits / wsram_width)
        k_partial_n_full_events = tiles.k_partial_n_full_total_tiles * math.ceil(tiles.k_partial_n_full_tile_bits / wsram_width)
        k_partial_n_partial_events = tiles.k_partial_n_partial_total_tiles * math.ceil(tiles.k_partial_n_partial_tile_bits / wsram_width)

        # M x N Output matrix osram <-> dram events (osram reads)
        m_full_n_full_events = tiles.m_full_n_full_total_tiles * tiles.k_tiles * math.ceil(tiles.m_full_n_full_tile_bits / osram_width)
        m_full_n_partial_events = tiles.m_full_n_partial_total_tiles * tiles.k_tiles * math.ceil(tiles.m_full_n_partial_tile_bits / osram_width)
        m_partial_n_full_events = tiles.m_partial_n_full_total_tiles * tiles.k_tiles * math.ceil(tiles.m_partial_n_full_tile_bits / osram_width)
        m_partial_n_partial_events = tiles.m_partial_n_partial_total_tiles * tiles.k_tiles * math.ceil(tiles.m_partial_n_partial_tile_bits / osram_width)

    # Total dram <-> sram events
    m_k_events = m_full_k_full_events + m_full_k_partial_events + m_partial_k_full_events + m_partial_k_partial_events
    k_n_events = k_full_n_full_events + k_full_n_partial_events + k_partial_n_full_events + k_partial_n_partial_events
    m_n_events = m_full_n_full_events + m_full_n_partial_events + m_partial_n_full_events + m_partial_n_partial_events

    isram_offchip_writes_dict = OrderedDict({'count': m_k_events})
    wsram_offchip_writes_dict = OrderedDict({'count': k_n_events})
    osram_offchip_reads_dict = OrderedDict({'count': m_n_events})
    osram_offchip_writes_dict = OrderedDict({'count': 0})
    dram_input_reads_dict = OrderedDict({'count': m_k_events})
    dram_weight_reads_dict = OrderedDict({'count': k_n_events})
    dram_output_reads_dict = OrderedDict({'count': 0})
    dram_output_writes_dict = OrderedDict({'count': m_n_events})

    performance_dict = OrderedDict()

    performance_dict['subevent'] = OrderedDict({
        'isram_offchip_writes': isram_offchip_writes_dict,
        'wsram_offchip_writes': wsram_offchip_writes_dict,
        'osram_offchip_reads': osram_offchip_reads_dict,
        'osram_offchip_writes': osram_offchip_writes_dict,
        'dram_input_reads': dram_input_reads_dict,
        'dram_weight_reads': dram_weight_reads_dict,
        'dram_output_reads': dram_output_reads_dict,
        'dram_output_writes': dram_output_writes_dict
    })

    return performance_dict

def onchip_gemm_scheduling(m, k, n, tiles: TiledGEMM, architecture_dict: OrderedDict, workload_dict: OrderedDict) -> list[TiledGEMM]:

    # Retrieve dicts
    arch = workload_dict['architecture']

    if arch == 'tensor':
        array_height = architecture_dict['wfifo']['instance'][-2]
        array_width = architecture_dict['ififo']['instance'][-2]
        array_depth = architecture_dict['wfifo']['instance'][-1]
    else:
        array_height = architecture_dict['wfifo']['instance'][-1]
        array_width = architecture_dict['ififo']['instance'][-1]
        array_depth = None

    tile_m = min(array_width, m)
    tile_k = min(array_height, k)
    tile_n = min(array_height, n)

    # Retrieve Reads/Writes for all possible tiling configurations
    tiling_configurations = []
    tiling_configurations.append(TiledGEMM(batch=tiles.m_full_k_full_n_full_total_tiles, m=tiles.tile_m, k=tiles.tile_k, n=tiles.tile_n, tile_m=tile_m, tile_k=tile_k, tile_n=tile_n, m_k_bitwidth=tiles.m_k_bitwidth, k_n_bitwidth=tiles.k_n_bitwidth, m_n_bitwidth=tiles.m_n_bitwidth, array_width=array_width, array_height=array_height, array_depth=array_depth))                                      # m full k full n full
    tiling_configurations.append(TiledGEMM(batch=tiles.m_full_k_full_n_partial_total_tiles, m=tiles.tile_m, k=tiles.tile_k, n=tiles.tile_n_partial, tile_m=tile_m, tile_k=tile_k, tile_n=tile_n, m_k_bitwidth=tiles.m_k_bitwidth, k_n_bitwidth=tiles.k_n_bitwidth, m_n_bitwidth=tiles.m_n_bitwidth, array_width=array_width, array_height=array_height, array_depth=array_depth))                           # m full k full n partial
    tiling_configurations.append(TiledGEMM(batch=tiles.m_full_k_partial_n_full_total_tiles, m=tiles.tile_m, k=tiles.tile_k_partial, n=tiles.tile_n, tile_m=tile_m, tile_k=tile_k, tile_n=tile_n, m_k_bitwidth=tiles.m_k_bitwidth, k_n_bitwidth=tiles.k_n_bitwidth, m_n_bitwidth=tiles.m_n_bitwidth, array_width=array_width, array_height=array_height, array_depth=array_depth))                           # m full k partial n full
    tiling_configurations.append(TiledGEMM(batch=tiles.m_full_k_partial_n_partial_total_tiles, m=tiles.tile_m, k=tiles.tile_k_partial, n=tiles.tile_n_partial, tile_m=tile_m, tile_k=tile_k, tile_n=tile_n, m_k_bitwidth=tiles.m_k_bitwidth, k_n_bitwidth=tiles.k_n_bitwidth, m_n_bitwidth=tiles.m_n_bitwidth, array_width=array_width, array_height=array_height, array_depth=array_depth))                # m full k partial n partial
    tiling_configurations.append(TiledGEMM(batch=tiles.m_partial_k_full_n_full_total_tiles, m=tiles.tile_m_partial, k=tiles.tile_k, n=tiles.tile_n, tile_m=tile_m, tile_k=tile_k, tile_n=tile_n, m_k_bitwidth=tiles.m_k_bitwidth, k_n_bitwidth=tiles.k_n_bitwidth, m_n_bitwidth=tiles.m_n_bitwidth, array_width=array_width, array_height=array_height, array_depth=array_depth))                           # m partial k full n full
    tiling_configurations.append(TiledGEMM(batch=tiles.m_partial_k_full_n_partial_total_tiles, m=tiles.tile_m_partial, k=tiles.tile_k, n=tiles.tile_n_partial, tile_m=tile_m, tile_k=tile_k, tile_n=tile_n, m_k_bitwidth=tiles.m_k_bitwidth, k_n_bitwidth=tiles.k_n_bitwidth, m_n_bitwidth=tiles.m_n_bitwidth, array_width=array_width, array_height=array_height, array_depth=array_depth))                # m partial k full n partial
    tiling_configurations.append(TiledGEMM(batch=tiles.m_partial_k_partial_n_full_total_tiles, m=tiles.tile_m_partial, k=tiles.tile_k_partial, n=tiles.tile_n, tile_m=tile_m, tile_k=tile_k, tile_n=tile_n, m_k_bitwidth=tiles.m_k_bitwidth, k_n_bitwidth=tiles.k_n_bitwidth, m_n_bitwidth=tiles.m_n_bitwidth, array_width=array_width, array_height=array_height, array_depth=array_depth))                # m partial k partial n full
    tiling_configurations.append(TiledGEMM(batch=tiles.m_partial_k_partial_n_partial_total_tiles, m=tiles.tile_m_partial, k=tiles.tile_k_partial, n=tiles.tile_n_partial, tile_m=tile_m, tile_k=tile_k, tile_n=tile_n, m_k_bitwidth=tiles.m_k_bitwidth, k_n_bitwidth=tiles.k_n_bitwidth, m_n_bitwidth=tiles.m_n_bitwidth, array_width=array_width, array_height=array_height, array_depth=array_depth))     # m partial k partial n partial
    partial_tiles_tiling = []

    for tiling in tiling_configurations:
        if tiling.is_valid:
            partial_tiles_tiling.append(tiling)

    if partial_tiles_tiling == []:
        raise ValueError('Invalid tiling configuration')

    return partial_tiles_tiling

def onchip_gemm_events(onchip_tiling: list[TiledGEMM], architecture_dict: OrderedDict, workload_dict: OrderedDict) -> OrderedDict:
    """
    Estimates events from all node srams <-> nodes for GEMM operations.
    """

    performance_dict = OrderedDict()

    isram_onchip_reads_count = 0
    wsram_onchip_reads_count = 0
    osram_onchip_reads_count = 0
    osram_onchip_writes_count = 0

    for tiling in onchip_tiling:
        if not tiling.is_valid:
            raise ValueError('Invalid tiling configuration')
        isram_reads, wsram_reads, osram_reads, osram_writes = onchip_gemm_tile_events(tiling, architecture_dict, workload_dict)
        isram_onchip_reads_count += isram_reads
        wsram_onchip_reads_count += wsram_reads
        osram_onchip_reads_count += osram_reads
        osram_onchip_writes_count += osram_writes

    isram_onchip_reads_dict = OrderedDict({'count': isram_onchip_reads_count})
    wsram_onchip_reads_dict = OrderedDict({'count': wsram_onchip_reads_count})
    osram_onchip_reads_dict = OrderedDict({'count': osram_onchip_reads_count})
    osram_onchip_writes_dict = OrderedDict({'count': osram_onchip_writes_count})

    performance_dict['subevent'] = OrderedDict({
        'isram_onchip_reads': isram_onchip_reads_dict,
        'wsram_onchip_reads': wsram_onchip_reads_dict,
        'osram_onchip_reads': osram_onchip_reads_dict,
        'osram_onchip_writes': osram_onchip_writes_dict
    })

    return performance_dict

def onchip_gemm_tile_events(tiles: TiledGEMM, architecture_dict: OrderedDict, workload_dict: OrderedDict) -> tuple:
    """
    Estimates events from node srams <-> node for GEMM operation.
    """

    # Retrieve dicts
    stationary = workload_dict['node_stationary']
    isram_width = architecture_dict['isram']['query']['width']
    wsram_width = architecture_dict['wsram']['query']['width']
    osram_width = architecture_dict['osram']['query']['width']

    # scheduling
    if stationary == 'os':
        # output stationary scheduling
        # M x K Input matrix isram -> node writes
        m_full_k_full_read_events = tiles.m_full_k_full_total_tiles * tiles.n_tiles * math.ceil(tiles.m_full_k_full_tile_bits / isram_width)
        m_full_k_partial_read_events = tiles.m_full_k_partial_total_tiles * tiles.n_tiles * math.ceil(tiles.m_full_k_partial_tile_bits / isram_width)
        m_partial_k_full_read_events = tiles.m_partial_k_full_total_tiles * tiles.n_tiles * math.ceil(tiles.m_partial_k_full_tile_bits / isram_width)
        m_partial_k_partial_read_events = tiles.m_partial_k_partial_total_tiles * tiles.n_tiles * math.ceil(tiles.m_partial_k_partial_tile_bits / isram_width)

        # K x N Weight matrix wsram -> node writes
        k_full_n_full_read_events = tiles.k_full_n_full_total_tiles * tiles.m_tiles * math.ceil(tiles.k_full_n_full_tile_bits / wsram_width)
        k_full_n_partial_read_events = tiles.k_full_n_partial_total_tiles * tiles.m_tiles * math.ceil(tiles.k_full_n_partial_tile_bits / wsram_width)
        k_partial_n_full_read_events = tiles.k_partial_n_full_total_tiles * tiles.m_tiles * math.ceil(tiles.k_partial_n_full_tile_bits / wsram_width)
        k_partial_n_partial_read_events = tiles.k_partial_n_partial_total_tiles * tiles.m_tiles * math.ceil(tiles.k_partial_n_partial_tile_bits / wsram_width)

        # M x N Output matrix osram -> node reads
        m_full_n_full_write_events = tiles.m_full_n_full_total_tiles * math.ceil(tiles.m_full_n_full_tile_bits / osram_width)
        m_full_n_partial_write_events = tiles.m_full_n_partial_total_tiles * math.ceil(tiles.m_full_n_partial_tile_bits / osram_width)
        m_partial_n_full_write_events = tiles.m_partial_n_full_total_tiles * math.ceil(tiles.m_partial_n_full_tile_bits / osram_width)
        m_partial_n_partial_write_events = tiles.m_partial_n_partial_total_tiles * math.ceil(tiles.m_partial_n_partial_tile_bits / osram_width)

        # M x N Output matrix osram -> node writes
        m_full_n_full_read_events = tiles.m_full_n_full_total_tiles * math.ceil(tiles.m_full_n_full_tile_bits / osram_width)
        m_full_n_partial_read_events = tiles.m_full_n_partial_total_tiles * math.ceil(tiles.m_full_n_partial_tile_bits / osram_width)
        m_partial_n_full_read_events = tiles.m_partial_n_full_total_tiles * math.ceil(tiles.m_partial_n_full_tile_bits / osram_width)
        m_partial_n_partial_read_events = tiles.m_partial_n_partial_total_tiles * math.ceil(tiles.m_partial_n_partial_tile_bits / osram_width)

    elif stationary == 'is':
        # input stationary scheduling
        # M x K Input matrix isram -> node writes
        m_full_k_full_read_events = tiles.m_full_k_full_total_tiles * math.ceil(tiles.m_full_k_full_tile_bits / isram_width)
        m_full_k_partial_read_events = tiles.m_full_k_partial_total_tiles * math.ceil(tiles.m_full_k_partial_tile_bits / isram_width)
        m_partial_k_full_read_events = tiles.m_partial_k_full_total_tiles * math.ceil(tiles.m_partial_k_full_tile_bits / isram_width)
        m_partial_k_partial_read_events = tiles.m_partial_k_partial_total_tiles * math.ceil(tiles.m_partial_k_partial_tile_bits / isram_width)

        # K x N Weight matrix wsram -> node writes
        k_full_n_full_read_events = tiles.k_full_n_full_total_tiles * tiles.m_tiles * math.ceil(tiles.k_full_n_full_tile_bits / wsram_width)
        k_full_n_partial_read_events = tiles.k_full_n_partial_total_tiles * tiles.m_tiles * math.ceil(tiles.k_full_n_partial_tile_bits / wsram_width)
        k_partial_n_full_read_events = tiles.k_partial_n_full_total_tiles * tiles.m_tiles * math.ceil(tiles.k_partial_n_full_tile_bits / wsram_width)
        k_partial_n_partial_read_events = tiles.k_partial_n_partial_total_tiles * tiles.m_tiles * math.ceil(tiles.k_partial_n_partial_tile_bits / wsram_width)

        # M x N Output matrix osram -> node reads
        m_full_n_full_write_events = tiles.m_full_n_full_total_tiles * tiles.k_tiles * math.ceil(tiles.m_full_n_full_tile_bits / osram_width)
        m_full_n_partial_write_events = tiles.m_full_n_partial_total_tiles * tiles.k_tiles * math.ceil(tiles.m_full_n_partial_tile_bits / osram_width)
        m_partial_n_full_write_events = tiles.m_partial_n_full_total_tiles * tiles.k_tiles * math.ceil(tiles.m_partial_n_full_tile_bits / osram_width)
        m_partial_n_partial_write_events = tiles.m_partial_n_partial_total_tiles * tiles.k_tiles * math.ceil(tiles.m_partial_n_partial_tile_bits / osram_width)

        # M x N Output matrix osram -> node writes
        m_full_n_full_read_events = tiles.m_full_n_full_total_tiles * (tiles.k_tiles - 1) * math.ceil(tiles.m_full_n_full_tile_bits / osram_width)
        m_full_n_partial_read_events = tiles.m_full_n_partial_total_tiles * (tiles.k_tiles - 1) * math.ceil(tiles.m_full_n_partial_tile_bits / osram_width)
        m_partial_n_full_read_events = tiles.m_partial_n_full_total_tiles * (tiles.k_tiles - 1) * math.ceil(tiles.m_partial_n_full_tile_bits / osram_width)
        m_partial_n_partial_read_events = tiles.m_partial_n_partial_total_tiles * (tiles.k_tiles - 1) * math.ceil(tiles.m_partial_n_partial_tile_bits / osram_width)

    elif stationary == 'ws':
        # weight stationary scheduling
        # M x K Input matrix isram -> node writes
        m_full_k_full_read_events = tiles.m_full_k_full_total_tiles * tiles.n_tiles * math.ceil(tiles.m_full_k_full_tile_bits / isram_width)
        m_full_k_partial_read_events = tiles.m_full_k_partial_total_tiles * tiles.n_tiles * math.ceil(tiles.m_full_k_partial_tile_bits / isram_width)
        m_partial_k_full_read_events = tiles.m_partial_k_full_total_tiles * tiles.n_tiles * math.ceil(tiles.m_partial_k_full_tile_bits / isram_width)
        m_partial_k_partial_read_events = tiles.m_partial_k_partial_total_tiles * tiles.n_tiles * math.ceil(tiles.m_partial_k_partial_tile_bits / isram_width)

        # K x N Weight matrix wsram -> node writes
        k_full_n_full_read_events = tiles.k_full_n_full_total_tiles * math.ceil(tiles.k_full_n_full_tile_bits / wsram_width)
        k_full_n_partial_read_events = tiles.k_full_n_partial_total_tiles * math.ceil(tiles.k_full_n_partial_tile_bits / wsram_width)
        k_partial_n_full_read_events = tiles.k_partial_n_full_total_tiles * math.ceil(tiles.k_partial_n_full_tile_bits / wsram_width)
        k_partial_n_partial_read_events = tiles.k_partial_n_partial_total_tiles * math.ceil(tiles.k_partial_n_partial_tile_bits / wsram_width)

        # M x N Output matrix osram -> node reads
        m_full_n_full_write_events = tiles.m_full_n_full_total_tiles * tiles.k_tiles * math.ceil(tiles.m_full_n_full_tile_bits / osram_width)
        m_full_n_partial_write_events = tiles.m_full_n_partial_total_tiles * tiles.k_tiles * math.ceil(tiles.m_full_n_partial_tile_bits / osram_width)
        m_partial_n_full_write_events = tiles.m_partial_n_full_total_tiles * tiles.k_tiles * math.ceil(tiles.m_partial_n_full_tile_bits / osram_width)
        m_partial_n_partial_write_events = tiles.m_partial_n_partial_total_tiles * tiles.k_tiles * math.ceil(tiles.m_partial_n_partial_tile_bits / osram_width)

        # M x N Output matrix osram -> node writes
        m_full_n_full_read_events = tiles.m_full_n_full_total_tiles * (tiles.k_tiles - 1) * math.ceil(tiles.m_full_n_full_tile_bits / osram_width)
        m_full_n_partial_read_events = tiles.m_full_n_partial_total_tiles * (tiles.k_tiles - 1) * math.ceil(tiles.m_full_n_partial_tile_bits / osram_width)
        m_partial_n_full_read_events = tiles.m_partial_n_full_total_tiles * (tiles.k_tiles - 1) * math.ceil(tiles.m_partial_n_full_tile_bits / osram_width)
        m_partial_n_partial_read_events = tiles.m_partial_n_partial_total_tiles * (tiles.k_tiles - 1) * math.ceil(tiles.m_partial_n_partial_tile_bits / osram_width)

    # Total sram <-> node events
    isram_reads = m_full_k_full_read_events + m_full_k_partial_read_events + m_partial_k_full_read_events + m_partial_k_partial_read_events
    wsram_reads = k_full_n_full_read_events + k_full_n_partial_read_events + k_partial_n_full_read_events + k_partial_n_partial_read_events
    osram_reads = m_full_n_full_read_events + m_full_n_partial_read_events + m_partial_n_full_read_events + m_partial_n_partial_read_events
    osram_writes = m_full_n_full_write_events + m_full_n_partial_write_events + m_partial_n_full_write_events + m_partial_n_partial_write_events

    return isram_reads, wsram_reads, osram_reads, osram_writes

def offchip_nonlinear_scheduling(batch, m, n, architecture_dict: OrderedDict, workload_dict: OrderedDict) -> TiledMatrix:
    """
    Sets tiling for dram <-> sram for single-node and multi-node configurations
    """
    # Retrieve dicts
    arch = workload_dict['architecture']
    wsram_dict = architecture_dict['wsram']
    osram_dict = architecture_dict['osram']
    input_bitwidth = workload_dict['activation_bitwidth']
    output_bitwidth = workload_dict['activation_bitwidth']
    if arch != 'tensor':
        array_width = architecture_dict['ififo']['instance'][-1]
        array_height = architecture_dict['wfifo']['instance'][-1]
    else:
        array_width = get_prod(architecture_dict['wfifo']['instance'][-2:])
        array_height = get_prod(architecture_dict['ififo']['instance'][-2:])
    architecture = workload_dict['architecture'].lower()

    # SRAM configurations
    # divide banks by to buffer
    wsram_bank = wsram_dict['query']['bank'] / 2
    osram_bank = osram_dict['query']['bank'] / 2

    wsram_width = wsram_dict['query']['width']
    osram_width = osram_dict['query']['width']

    wsram_depth = wsram_dict['query']['depth']
    osram_depth = osram_dict['query']['depth']

    wsram_size = wsram_bank * wsram_width * wsram_depth
    osram_size = osram_bank * osram_width * osram_depth

    wsram_elements = wsram_size / input_bitwidth
    osram_elements = osram_size / output_bitwidth

    # number of nodes
    nodes = wsram_dict['instance'][0] * wsram_dict['instance'][1] if 'router' in architecture_dict else 1

    # input and output matrices are the same size, so so input and output must fit in both
    if architecture == 'mugi':
        elements = min(wsram_elements, osram_elements)
    else:
        elements = osram_elements
    
    # initialize m tile size to m (in llms, very oftem batch size or smaller)
    # initialize n tile size to maximum size that fully utalizes noc
    tile_m = min(m, array_height)
    tile_n = ((n * batch) / nodes)

    # if tile_n does not fit in osram, reduce tile_n
    while(tile_n * tile_m > elements):
        tile_n /= 2

    # if small tile n, and is smaller than array, increase tile_n
    while((tile_n * tile_m * 2 < elements)):
        tile_n *= 2

    # if small tile n, and can increase m, increase tile_m
    # multiply by 2 to check if you can increase tile_m
    while((tile_m * tile_n * 2 < elements) and (tile_m < m)):
        tile_m *= 2

    if tile_m * tile_n > wsram_elements:
        raise ValueError('Tile size exceeds wsram size')
    if tile_m * tile_n > osram_elements:
        raise ValueError('Tile size exceeds osram size')
    
    tile_m = int(tile_m)
    tile_n = int(tile_n)

    

    tiles = TiledMatrix(batch=batch, m=m, n=n, tile_m=tile_m, tile_n=tile_n, m_n_bitwidth=input_bitwidth)

    if not tiles.is_valid:
        raise ValueError('Invalid tiling configuration')

    return tiles

def offchip_nonlinear_events(tiles: TiledMatrix, architecture_dict: OrderedDict, workload_dict: OrderedDict) -> OrderedDict:
    """
    Estimates dram and sram reads and writes to from offchip memory to onchip memory for non-linear
    """

    performance_dict = OrderedDict()

    # Retrieve dicts
    isram_width = architecture_dict['isram']['query']['width']
    osram_width = architecture_dict['osram']['query']['width']
    if 'irouter' in architecture_dict:
        router_height = architecture_dict['irouter']['instance'][0]
        router_width = architecture_dict['irouter']['instance'][1]
    else:
        router_height = 1
        router_width = 1
    architecture = workload_dict['architecture'].lower()

    noc_dim = router_height * router_width if 'router' in architecture_dict else 1

    m_full_n_full_events = tiles.m_full_n_full_total_tiles * math.ceil(tiles.m_full_n_full_tile_bits / osram_width)
    m_full_n_partial_events = tiles.m_full_n_partial_total_tiles * math.ceil(tiles.m_full_n_partial_tile_bits / osram_width)
    m_partial_n_full_events = tiles.m_partial_n_full_total_tiles * math.ceil(tiles.m_partial_n_full_tile_bits / osram_width)
    m_partial_n_partial_events = tiles.m_partial_n_partial_total_tiles * math.ceil(tiles.m_partial_n_partial_tile_bits / osram_width)

    m_n_event = m_full_n_full_events + m_full_n_partial_events + m_partial_n_full_events + m_partial_n_partial_events

    # If virtual lut is used in isram
    if architecture == 'mugi':
        lut_height = workload_dict['lut_height']
        lut_width = workload_dict['lut_width']
        weight_bitwidth = workload_dict['weight_bitwidth']
        lut_dim = lut_height * lut_width
        lut_bits = lut_dim * weight_bitwidth
        
        isram_writes_count = min(tiles.m_n_matrix_tiles, noc_dim) * math.ceil(lut_bits / isram_width)
        dram_input_reads_count = math.ceil(lut_bits / isram_width)
    else:
        isram_writes_count = 0
        dram_input_reads_count = 0

    wsram_writes_count = 0
    osram_reads_count = m_n_event
    osram_writes_count = m_n_event
    dram_weight_reads_count = 0
    dram_output_reads_count = m_n_event
    dram_output_writes_count = m_n_event

    isram_offchip_writes_dict = OrderedDict({'count': isram_writes_count})
    wsram_offchip_writes_dict = OrderedDict({'count': wsram_writes_count})
    osram_offchip_reads_dict = OrderedDict({'count': osram_reads_count})
    osram_offchip_writes_dict = OrderedDict({'count': osram_writes_count})
    dram_input_reads_dict = OrderedDict({'count': dram_input_reads_count})
    dram_weight_reads_dict = OrderedDict({'count': dram_weight_reads_count})
    dram_output_reads_dict = OrderedDict({'count': dram_output_reads_count})
    dram_output_writes_dict = OrderedDict({'count': dram_output_writes_count})

    performance_dict = OrderedDict()

    performance_dict['subevent'] = OrderedDict({
        'isram_offchip_writes': isram_offchip_writes_dict,
        'wsram_offchip_writes': wsram_offchip_writes_dict,
        'osram_offchip_reads': osram_offchip_reads_dict,
        'osram_offchip_writes': osram_offchip_writes_dict,
        'dram_input_reads': dram_input_reads_dict,
        'dram_weight_reads': dram_weight_reads_dict,
        'dram_output_reads': dram_output_reads_dict,
        'dram_output_writes': dram_output_writes_dict
    })

    return performance_dict

def onchip_nonlinear_events(function: str, onchip_tiling: list[TiledMatrix], architecture_dict: OrderedDict, workload_dict: OrderedDict) -> OrderedDict:
    """
    Estimates events from node srams <-> node for nonlinear operation.
    """

    performance_dict = OrderedDict()

    isram_onchip_reads_count = 0
    wsram_onchip_reads_count = 0
    osram_onchip_reads_count = 0
    osram_onchip_writes_count = 0

    for tiling in onchip_tiling:
        if not tiling.is_valid:
            raise ValueError('Invalid tiling configuration')
        isram_reads, wsram_reads, osram_reads, osram_writes = onchip_nonlinear_tile_events(function=function, tiles=tiling, architecture_dict=architecture_dict, workload_dict=workload_dict)
        isram_onchip_reads_count += isram_reads
        wsram_onchip_reads_count += wsram_reads
        osram_onchip_reads_count += osram_reads
        osram_onchip_writes_count += osram_writes

    isram_onchip_reads_dict = OrderedDict({'count': isram_onchip_reads_count})
    wsram_onchip_reads_dict = OrderedDict({'count': wsram_onchip_reads_count})
    osram_onchip_reads_dict = OrderedDict({'count': osram_onchip_reads_count})
    osram_onchip_writes_dict = OrderedDict({'count': osram_onchip_writes_count})

    performance_dict['subevent'] = OrderedDict({
        'isram_onchip_reads': isram_onchip_reads_dict,
        'wsram_onchip_reads': wsram_onchip_reads_dict,
        'osram_onchip_reads': osram_onchip_reads_dict,
        'osram_onchip_writes': osram_onchip_writes_dict
    })

    return performance_dict

def onchip_nonlinear_scheduling(m, n, offchip_tiles: TiledMatrix, architecture_dict: OrderedDict, workload_dict: OrderedDict) -> TiledMatrix:

    arch = workload_dict['architecture']
    if arch != 'tensor':
        wfifo_width = architecture_dict['wfifo']['instance'][-1]
        ofifo_width = architecture_dict['ofifo']['instance'][-1]
    else:
        wfifo_width = get_prod(architecture_dict['ififo']['instance'][-2:])
        ofifo_width = get_prod(architecture_dict['ofifo']['instance'][-2:])
    
    array_height = ofifo_width

    if wfifo_width != ofifo_width:
        raise ValueError('wfifo and ofifo width must be the same')

    tile_m = min(array_height, m)
    tile_n = 1

    tiling_configurations = []
    tiling_configurations.append(TiledMatrix(batch=offchip_tiles.m_full_n_full_total_tiles, m=offchip_tiles.tile_m, n=offchip_tiles.tile_n, tile_m=tile_m, tile_n=tile_n, m_n_bitwidth=offchip_tiles.m_n_bitwidth, array_height=array_height))                         # m full n full
    tiling_configurations.append(TiledMatrix(batch=offchip_tiles.m_full_n_partial_total_tiles, m=offchip_tiles.tile_m, n=offchip_tiles.tile_n_partial, tile_m=tile_m, tile_n=tile_n, m_n_bitwidth=offchip_tiles.m_n_bitwidth, array_height=array_height))              # m full n partial
    tiling_configurations.append(TiledMatrix(batch=offchip_tiles.m_partial_n_full_total_tiles, m=offchip_tiles.tile_m_partial, n=offchip_tiles.tile_n, tile_m=tile_m, tile_n=tile_n, m_n_bitwidth=offchip_tiles.m_n_bitwidth, array_height=array_height))              # m partial n full
    tiling_configurations.append(TiledMatrix(batch=offchip_tiles.m_partial_n_partial_total_tiles, m=offchip_tiles.tile_m_partial, n=offchip_tiles.tile_n_partial, tile_m=tile_m, tile_n=tile_n, m_n_bitwidth=offchip_tiles.m_n_bitwidth, array_height=array_height))   # m partial n partial

    partial_tiles_tiling = []

    for tiling in tiling_configurations:
        if tiling.is_valid:
            partial_tiles_tiling.append(tiling)

    if partial_tiles_tiling == []:
        raise ValueError('Invalid tiling configuration')

    return partial_tiles_tiling

def onchip_nonlinear_tile_events(function: str, tiles: TiledMatrix, architecture_dict: OrderedDict, workload_dict: OrderedDict) -> tuple:
    """
    Estimates events from node srams <-> node for nonlinear operation.
    """

    # check if LUT implementation (Mugi)
    architecture = workload_dict['architecture'].lower()

    # retrieve dicts
    isram_width = architecture_dict['isram']['query']['width']
    wsram_width = architecture_dict['wsram']['query']['width']
    osram_width = architecture_dict['osram']['query']['width']

    m_full_n_full_events = tiles.m_full_n_full_total_tiles * math.ceil(tiles.m_full_n_full_tile_bits / osram_width)
    m_full_n_partial_events = tiles.m_full_n_partial_total_tiles * math.ceil(tiles.m_full_n_partial_tile_bits / osram_width)
    m_partial_n_full_events = tiles.m_partial_n_full_total_tiles * math.ceil(tiles.m_partial_n_full_tile_bits / osram_width)
    m_partial_n_partial_events = tiles.m_partial_n_partial_total_tiles * math.ceil(tiles.m_partial_n_partial_tile_bits / osram_width)

    m_n_events = m_full_n_full_events + m_full_n_partial_events + m_partial_n_full_events + m_partial_n_partial_events

    if function == 'softmax':
        osram_reads = m_n_events * 2
        osram_writes = m_n_events * 2
    else: # default mapping for activation function. In this case, SiLU. Holds true to activation functions that apply element-wise operations (unlike softmax which divides by sum).
        osram_reads = m_n_events
        osram_writes = m_n_events

    if architecture == 'mugi':
        lut_height = workload_dict['lut_height']
        window_width = workload_dict['window_width']
        weight_bitwidth = workload_dict['weight_bitwidth']
        lut_bits = window_width * weight_bitwidth

        isram_reads = (tiles.m_n_total_tiles * lut_height) * math.ceil(lut_bits / isram_width)
    else:
        isram_reads = 0

    wsram_reads = 0

    return isram_reads, wsram_reads, osram_reads, osram_writes


from collections import OrderedDict
from archx.utils import get_prod

def input_nonlinear(architecture_dict: OrderedDict, workload_dict: OrderedDict=None)->OrderedDict: 
    performance_dict = OrderedDict()
    workload_dict = next(iter(workload_dict.values()))['configuration']
    router_dim = get_prod(architecture_dict['irouter']['instance']) if 'irouter' in architecture_dict else 1

    frequency = architecture_dict['ififo']['query']['frequency']

    ififo_dim = architecture_dict['ififo']['instance'][-1]
    lut_height = workload_dict['lut_height']

    cycle_count = lut_height / router_dim
    performance_dict['cycle_count'] = OrderedDict({'value': cycle_count, 'unit': 'cycle'})
    performance_dict['runtime'] = OrderedDict({'value': cycle_count / 1000 / frequency, 'unit': 'ms'})

    ififo_dict = OrderedDict({'count': ififo_dim * lut_height})

    performance_dict['subevent'] = OrderedDict({'ififo': ififo_dict})
    return performance_dict

def weight_nonlinear(architecture_dict: OrderedDict, workload_dict: OrderedDict=None)->OrderedDict:
    performance_dict = OrderedDict()
    router_dim = get_prod(architecture_dict['irouter']['instance']) if 'irouter' in architecture_dict else 1
    workload_dict = next(iter(workload_dict.values()))['configuration']
    lut_height = workload_dict['lut_height']
    frequency = architecture_dict['round']['query']['frequency']

    round_dim = architecture_dict['round']['instance'][-1]
    sign_mantissa_register_dim = architecture_dict['sign_mantissa_register']['instance'][-1]
    exponent_register_dim = architecture_dict['exponent_register']['instance'][-1]
    exp_clamp_dim = architecture_dict['exp_clamp']['instance'][-1]
    wfifo_dim = architecture_dict['wfifo']['instance'][-1]
    max_tree_dim = architecture_dict['max_tree']['instance'][-1]
    max_tree_register_dim = architecture_dict['max_tree_register']['instance'][-1]
    magnitude_register_dim = architecture_dict['magnitude_register']['instance'][-1]
    sign_register_dim = architecture_dict['sign_register']['instance'][-1]
    sign_fifo_dim = architecture_dict['sign_fifo']['instance'][-1]
    post_comp_dim = architecture_dict['post_comp']['instance'][-1]
    ofifo_dim = architecture_dict['ofifo']['instance'][-1]

    if 'window_select' in architecture_dict:
        window_select_dim = architecture_dict['window_select']['instance'][-1]
    if 'lut_register' in architecture_dict:
        lut_register_dim = architecture_dict['lut_register']['instance'][-2] * architecture_dict['lut_register']['instance'][-4]
    if 'lut_decoder' in architecture_dict:
        lut_decoder_dim = architecture_dict['lut_decoder']['instance'][-1]

    cycle_count = lut_height / router_dim
    performance_dict['cycle_count'] = OrderedDict({'value': cycle_count, 'unit': 'cycle'})
    performance_dict['runtime'] = OrderedDict({'value': cycle_count / 1000 / frequency, 'unit': 'ms'})

    round_dict = OrderedDict({'count': round_dim})
    sign_mantissa_register_dict = OrderedDict({'count': sign_mantissa_register_dim})
    exponent_register_dict = OrderedDict({'count': exponent_register_dim})
    exp_clamp_dict = OrderedDict({'count': exp_clamp_dim})
    wfifo_dict = OrderedDict({'count': wfifo_dim})
    max_tree_dict = OrderedDict({'count': max_tree_dim})
    max_tree_register_dict = OrderedDict({'count': max_tree_register_dim})
    magnitude_register_dict = OrderedDict({'count': magnitude_register_dim})
    sign_register_dict = OrderedDict({'count': sign_register_dim})
    sign_fifo_dict = OrderedDict({'count': sign_fifo_dim})
    post_comp_dict = OrderedDict({'count': post_comp_dim})
    ofifo_dict = OrderedDict({'count': ofifo_dim})

    if 'window_select' in architecture_dict:
        window_select_dict = OrderedDict({'count': window_select_dim})
    if 'lut_register' in architecture_dict:
        lut_register_dict = OrderedDict({'count': lut_register_dim})
    if 'lut_decoder' in architecture_dict:
        lut_decoder_dict = OrderedDict({'count': lut_decoder_dim * lut_height})

    performance_dict['subevent'] = OrderedDict({'round': round_dict,
                                                'sign_mantissa_register': sign_mantissa_register_dict,
                                                'exponent_register': exponent_register_dict,
                                                'exp_clamp': exp_clamp_dict,
                                                'wfifo': wfifo_dict,
                                                'max_tree': max_tree_dict,
                                                'max_tree_register': max_tree_register_dict,
                                                'magnitude_register': magnitude_register_dict,
                                                'sign_register': sign_register_dict,
                                                'sign_fifo': sign_fifo_dict,
                                                'post_comp': post_comp_dict,
                                                'ofifo': ofifo_dict})
    
    if 'window_select' in architecture_dict:
        performance_dict['subevent']['window_select'] = window_select_dict
    if 'lut_register' in architecture_dict:
        performance_dict['subevent']['lut_register'] = lut_register_dict
    if 'lut_decoder' in architecture_dict:
        performance_dict['subevent']['lut_decoder'] = lut_decoder_dict

    return performance_dict

def weight_reuse_nonlinear(architecture_dict: OrderedDict, workload_dict: OrderedDict=None)->OrderedDict:
    performance_dict = OrderedDict()
    workload_dict = next(iter(workload_dict.values()))['configuration']
    frequency = architecture_dict['comparator']['query']['frequency']
    lut_height = workload_dict['lut_height']
    router_dim = get_prod(architecture_dict['irouter']['instance']) if 'irouter' in architecture_dict else 1

    comparator_dim = architecture_dict['comparator']['instance'][-1]
    
    cycle_count = lut_height / router_dim
    performance_dict['cycle_count'] = OrderedDict({'value': cycle_count, 'unit': 'cycle'})
    performance_dict['runtime'] = OrderedDict({'value': cycle_count / 1000 / frequency, 'unit': 'ms'})

    comparator_dict = OrderedDict({'count': comparator_dim * lut_height})

    performance_dict['subevent'] = OrderedDict({'comparator': comparator_dict})
    return performance_dict

def array_nonlinear(architecture_dict: OrderedDict, workload_dict: OrderedDict=None)->OrderedDict:
    performance_dict = OrderedDict()
    router_dim = get_prod(architecture_dict['irouter']['instance']) if 'irouter' in architecture_dict else 1

    frequency = architecture_dict['temporal_register']['query']['frequency']
    cycles = next(iter(workload_dict.values()))['configuration']['cycles']

    temporal_register_dim = get_prod(architecture_dict['temporal_register']['instance'][-2:])
    and_gate_dim = get_prod(architecture_dict['and_gate']['instance'][-2:])
    or_gate_dim = get_prod(architecture_dict['or_gate']['instance'][-3:-1])

    cycle_count = cycles / router_dim
    performance_dict['cycle_count'] = OrderedDict({'value': cycle_count, 'unit': 'cycle'})
    performance_dict['runtime'] = OrderedDict({'value': cycle_count / 1000 / frequency, 'unit': 'ms'})

    temporal_register_dict = OrderedDict({'count': temporal_register_dim})
    and_gate_dict = OrderedDict({'count': and_gate_dim})
    or_gate_dict = OrderedDict({'count': or_gate_dim})

    performance_dict['subevent'] = OrderedDict({'temporal_register': temporal_register_dict,
                                                'and_gate': and_gate_dict,
                                                'or_gate': or_gate_dict})
    return performance_dict

def array_fifo_nonlinear(architecture_dict: OrderedDict, workload_dict: OrderedDict=None)->OrderedDict:
    performance_dict = OrderedDict()
    router_dim = get_prod(architecture_dict['irouter']['instance']) if 'irouter' in architecture_dict else 1
    cycles = next(iter(workload_dict.values()))['configuration']['cycles']

    frequency = architecture_dict['pe_fifo']['query']['frequency']

    or_gate_dim = get_prod(architecture_dict['or_gate']['instance'][-3:1])
    pe_fifo_dim = architecture_dict['pe_fifo']['instance'][-1]

    cycle_count = cycles / router_dim
    performance_dict['cycle_count'] = OrderedDict({'value': cycle_count, 'unit': 'cycle'})
    performance_dict['runtime'] = OrderedDict({'value': cycle_count / 1000 / frequency, 'unit': 'ms'})

    or_gate_dict = OrderedDict({'count': or_gate_dim})
    pe_fifo_dict = OrderedDict({'count': pe_fifo_dim})

    performance_dict['subevent'] = OrderedDict({'or_gate': or_gate_dict,
                                                'pe_fifo': pe_fifo_dict})
    return performance_dict

def summation(architecture_dict: OrderedDict, workload_dict: OrderedDict=None)->OrderedDict:
    performance_dict = OrderedDict()
    router_dim = get_prod(architecture_dict['irouter']['instance']) if 'irouter' in architecture_dict else 1
    cycles = next(iter(workload_dict.values()))['configuration']['cycles']

    frequency = architecture_dict['adder']['query']['frequency']

    adder_dim = architecture_dict['adder']['instance'][-1]

    cycle_count = cycles / router_dim
    performance_dict['cycle_count'] = OrderedDict({'value': cycle_count, 'unit': 'cycle'})
    performance_dict['runtime'] = OrderedDict({'value': cycle_count / 1000 / frequency, 'unit': 'ms'})

    adder_dict = OrderedDict({'count': adder_dim})

    performance_dict['subevent'] = OrderedDict({'adder': adder_dict})
    return performance_dict

def gemm_nonlinear(architecture_dict: OrderedDict, workload_dict: OrderedDict=None)->OrderedDict:
    performance_dict = OrderedDict()

    performance_dict['cycle_count'] = OrderedDict({'value': 0, 'unit': 'cycle'})
    performance_dict['runtime'] = OrderedDict({'value': 0, 'unit': 'ms'})

    multiplier_dict = OrderedDict({'count': 0})
    multiplier_register_dict = OrderedDict({'count': 0})
    sign_register_dim = OrderedDict({'count': 0})
    sign_fifo_dim = OrderedDict({'count': 0})
    sign_xor_dict = OrderedDict({'count': 0})

    performance_dict['subevent'] = OrderedDict({'multiplier': multiplier_dict,
                                                'multiplier_register': multiplier_register_dict,
                                                'sign_register': sign_register_dim,
                                                'sign_fifo': sign_fifo_dim,
                                                'sign_xor': sign_xor_dict})
    return performance_dict




from zoo.llm.results.query.utils import query_performance_nonlinear_metrics, compute_throughput_efficiancy, load_yaml, geomean
import pandas as pd
import os
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors

def query(input_path, output_path):
    vlp_list = ['mugi', 'carat']
    vlp_arch_dim_list = ['128x8', '256x8']

    mugi_subarch_list = ['vlp']

    baseline_list = ['systolic']
    baseline_arch_dim_list = ['16x16']
    baseline_subarch_list = ['mac', 'pwl', 'taylor']

    mugi_throughput_module = 'magnitude_register'
    baseline_throughput_module = 'accumulator_vector'
    approximate_throughtput_module = 'adder_vector'

    model_list = ['llama_2_7b', 'llama_2_13b', 'llama_2_70b']
    max_seq_len_list = ['max_seq_len_128', 'max_seq_len_256', 'max_seq_len_512', 'max_seq_len_1024', 'max_seq_len_2048', 'max_seq_len_4096']
    batch_size = 'batch_size_8'
    network = 'single_node'

    nonlinear_breakdown_df = pd.DataFrame()

    for arch in vlp_list + baseline_list + ['tensor']:
       for arch_dim in (vlp_arch_dim_list if arch in vlp_list else baseline_arch_dim_list if arch in baseline_list else ['8x16x16'] if arch in ['tensor'] else ['']):
            for subarch in (baseline_subarch_list if arch in baseline_list else mugi_subarch_list if arch in ['mugi'] else ['']):
                if arch == 'simd' and subarch in ['pwl', 'taylor']:
                    continue
                for max_seq_len in max_seq_len_list:
                    softmax_list = []
                    silu_list = []
                    for model in model_list:
                        module = mugi_throughput_module if arch == 'mugi' else approximate_throughtput_module if subarch in ['pwl', 'taylor'] else baseline_throughput_module
                        termination_path = 'full_termination' if arch == 'mugi' else ''
                        run_path = os.path.normpath(f'{input_path}{arch}/{network}/{subarch}/{arch_dim}/{model}/{max_seq_len}/{batch_size}/{termination_path}/')
                        yaml_dict = load_yaml(run_path)

                        event_graph = yaml_dict['event_graph']
                        metric_dict = yaml_dict['metric_dict']
                        sm_metric_dict = query_performance_nonlinear_metrics(event_graph=event_graph, metric_dict=metric_dict, module=module, workload=model, event = 'softmax')
                        silu_metric_dict = query_performance_nonlinear_metrics(event_graph=event_graph, metric_dict=metric_dict, module=module, workload=model, event = 'silu')

                        if subarch == 'taylor':
                            sm_metric_dict['flops'] /= 9

                        silu_throughput_eff_dict = compute_throughput_efficiancy(silu_metric_dict)
                        sm_throughput_eff_dict = compute_throughput_efficiancy(sm_metric_dict)

                        softmax_list.append(sm_throughput_eff_dict)
                        silu_list.append(silu_throughput_eff_dict)

                    sm_throughput_eff_dict = geomean(softmax_list)
                    silu_throughput_eff_dict = geomean(silu_list)

                    sm_metric_df = pd.DataFrame(sm_throughput_eff_dict, index=[0])
                    sm_metric_df['function'] = 'softmax'
                    silu_metric_df = pd.DataFrame(silu_throughput_eff_dict, index=[0])
                    silu_metric_df['function'] = 'silu'

                    nonlinear_metric_df = pd.concat([sm_metric_df, silu_metric_df])
                    nonlinear_metric_df['arch'] = arch
                    nonlinear_metric_df['subarch'] = subarch
                    nonlinear_metric_df['arch_dim'] = arch_dim
                    nonlinear_metric_df['max_seq_len'] = max_seq_len

                    nonlinear_metric_df = nonlinear_metric_df.drop(columns=['power', 'energy'], errors='ignore')
                    nonlinear_breakdown_df = pd.concat([nonlinear_breakdown_df, nonlinear_metric_df], axis=0)

    nonlinear_breakdown_df.to_csv(output_path + 'nonlinear_breakdown.csv', index=False)

    baseline_df = nonlinear_breakdown_df[
        (nonlinear_breakdown_df['arch'] == 'systolic') &
        (nonlinear_breakdown_df['subarch'] == 'mac') &
        (nonlinear_breakdown_df['arch_dim'] == '16x16')
    ]

    numeric_columns = baseline_df.select_dtypes(include=['number']).columns
    columns_to_merge = ['max_seq_len', 'function'] + list(numeric_columns)

    merged_df = nonlinear_breakdown_df.merge(
        baseline_df[columns_to_merge],
        on=['max_seq_len', 'function'],
        suffixes=('', '_baseline')
    )

    merged_df['throughput'] = merged_df['throughput'] / merged_df['throughput_baseline']
    merged_df['energy_efficiency'] = merged_df['energy_efficiency'] / merged_df['energy_efficiency_baseline']
    merged_df['power_efficiency'] = merged_df['power_efficiency'] / merged_df['power_efficiency_baseline']

    normalized_df = merged_df.drop(
        columns=['throughput_baseline', 'energy_efficiency_baseline', 'power_efficiency_baseline']
    )

    normalized_df.to_csv(output_path + 'nonlinear_breakdown_norm.csv', index=False)

def lighten_color(color, amount=0.5):
    try:
        c = mcolors.cnames[color]
    except KeyError:
        c = color
    c = mcolors.ColorConverter.to_rgb(c)
    return [(1 - amount) * x + amount for x in c]

def figure(input_path: str, output_path: str):
    """Generate nonlinear breakdown figure directly from CSV data."""
    try:
        # Read the normalized CSV data
        data_df = pd.read_csv(input_path + 'nonlinear_breakdown_norm.csv')
    except FileNotFoundError:
        print(f"Error: Could not find CSV file at {input_path}nonlinear_breakdown_norm.csv")
        return
    except Exception as e:
        print(f"Error reading CSV: {e}")
        return

    # Extract unique sequence lengths and sort them
    seq_lens = sorted(data_df['max_seq_len'].str.replace('max_seq_len_', '').astype(int).unique())
    x_ticks = [str(seq_len) for seq_len in seq_lens]
    print(f"Available sequence lengths: {x_ticks}")
    
    # Focus on specific sequence lengths for cleaner visualization
    target_seq_lens = ['128', '256', '512', '1024', '2048', '4096']
    data_dict = {'XTicks': target_seq_lens}
    
    # Architecture mapping
    arch_mapping = {
        'mugi': 'Mugi',
        'carat': 'Carat', 
        'systolic': 'VA',
        'tensor': 'Tensor'
    }
    
    # Create legend entries by extracting data from CSV
    for _, row in data_df.iterrows():
        arch = arch_mapping.get(row['arch'], row['arch'])
        function = row['function']
        arch_dim = row['arch_dim']
        subarch = row['subarch'] if pd.notna(row['subarch']) else ''
        seq_len = row['max_seq_len'].replace('max_seq_len_', '')
        
        # Skip entries not in our target sequence lengths
        if seq_len not in target_seq_lens:
            continue
            
        # Create key for this configuration
        function_key = 'SM ' if function == 'softmax' else 'SiLU '
        
        # Extract dimension info
        if arch in ['Mugi', 'Carat']:
            dim_size = arch_dim.split('x')[0]
        elif arch == 'Tensor':
            dim_size = '16'  # Use 16 for tensor
        else:  # VA/systolic
            dim_size = arch_dim.split('x')[0]
            
        dim_key = f' ({dim_size})'
        
        # Handle subarch
        subarch_key = ''
        if subarch and subarch != 'mac':
            if subarch == 'vlp':
                subarch_key = ''  # vlp is default for Mugi
            elif subarch == 'lut':
                subarch_key = ' LUT'
            elif subarch == 'pwl':
                subarch_key = ' PWL'
            elif subarch == 'taylor':
                subarch_key = ' Taylor'
                
        # Skip Taylor for SiLU (not commonly used)
        if subarch == 'taylor' and function == 'silu':
            continue
            
        # Skip PWL/Taylor for non-VA architectures (except Tensor which doesn't use them)
        if arch not in ['VA', 'Tensor'] and subarch in ['pwl', 'taylor']:
            continue
            
        key = function_key + arch + subarch_key + dim_key
        
        # Initialize data structure if not exists
        if key not in data_dict:
            data_dict[key] = {
                'NormThroughput': [0] * len(target_seq_lens),
                'NormEnergyEfficiency': [0] * len(target_seq_lens),
                'NormPowerEfficiency': [0] * len(target_seq_lens)
            }
            
        # Find index for this sequence length
        try:
            seq_idx = target_seq_lens.index(seq_len)
            data_dict[key]['NormThroughput'][seq_idx] = row['throughput']
            data_dict[key]['NormEnergyEfficiency'][seq_idx] = row['energy_efficiency'] 
            data_dict[key]['NormPowerEfficiency'][seq_idx] = row['power_efficiency']
        except ValueError:
            continue  # Skip if sequence length not in our targets
    
    # Remove any empty entries
    data_dict = {k: v for k, v in data_dict.items() if k != 'XTicks' and any(v['NormThroughput'])}
    
    # Add XTicks for sequence lengths
    data_dict['XTicks'] = target_seq_lens
    
    print(f"Generated {len(data_dict) - 1} data series:")
    for key in data_dict:
        if key != 'XTicks':
            print(f"  {key}")
    
    data = data_dict

    # -------------------------------
    # 2) FIGURE AND FONT SETTINGS
    # -------------------------------
    fig_width_pt = 240  # ACM single-column width in points
    fig_width = fig_width_pt / 72  # inches
    fig_height = fig_width / 1.8  # Adjusted height for readability

    font_title = 7
    font_tick = 6

    # -------------------------------
    # 3) COLOR SCHEME
    # -------------------------------
    base_colors = {
        'Mugi': "#0B752B",
        'Carat': "#602696", 
        'VA': "#0F6EA5",
        'PWL': "#C7A612",
        'Taylor': "#C21A1A",
        'Tensor': "#2A9B8E"  
    }

    # Generate colors for each data series
    colors = {}
    for key in data.keys():
        if 'Mugi' in key:
            if '128' in key:
                colors[key] = lighten_color(base_colors['Mugi'], 0.4)
            else:
                colors[key] = base_colors['Mugi']
        elif 'Carat' in key:
            if '128' in key:
                colors[key] = lighten_color(base_colors['Carat'], 0.4)
            else:
                colors[key] = base_colors['Carat']
        elif 'VA' in key:
            if 'PWL' in key:
                if 'SM' in key:
                    colors[key] = lighten_color(base_colors['PWL'], 0.4)
                else:
                    colors[key] = base_colors['PWL']
            elif 'Taylor' in key:
                colors[key] = base_colors['Taylor']
            else:
                if 'SM' in key:
                    colors[key] = lighten_color(base_colors['VA'], 0.4)
                else:
                    colors[key] = base_colors['VA']
        elif 'Tensor' in key:
            if 'SM' in key:
                colors[key] = lighten_color(base_colors['Tensor'], 0.4)
            else:
                colors[key] = base_colors['Tensor']


        else:
            colors[key] = 'black'  # fallback

    # -------------------------------
    # 4) CREATE SUBPLOTS AND PLOT AS BAR CHART
    # -------------------------------
    fig, (ax1, ax2, ax3) = plt.subplots(3, 1, figsize=(fig_width, fig_height), sharex=True)

    x_labels = data['XTicks']
    x = np.arange(len(x_labels))
    
    # Get all categories (skip 'XTicks')
    categories = [key for key in data.keys() if key != 'XTicks']
    num_categories = len(categories)
    bar_width = 0.93 / num_categories  # Slightly narrower bars for better spacing

    for i, key in enumerate(categories):
        throughput = data[key]['NormThroughput']
        energy_eff = data[key]['NormEnergyEfficiency']
        power_eff = data[key]['NormPowerEfficiency']
        
        color = colors.get(key, 'black')
        
        # Extract subarch for legend labels
        legend_label = key
        
        # Calculate bar positions
        bar_positions = x + (i - num_categories/2 + 0.5) * bar_width
        
        ax1.bar(bar_positions, throughput, width=bar_width, color=color, label=legend_label, 
                alpha=0.8, edgecolor='black', linewidth=0.15)
        ax2.bar(bar_positions, energy_eff, width=bar_width, color=color, label=legend_label,
                alpha=0.8, edgecolor='black', linewidth=0.15)
        ax3.bar(bar_positions, power_eff, width=bar_width, color=color, label=legend_label,
                alpha=0.8, edgecolor='black', linewidth=0.15)

    # -------------------------------
    # 6) FORMAT SUBPLOTS AND LEGEND
    # -------------------------------
    for ax, title in zip([ax1, ax2, ax3], ['Norm Throughput', 'Norm Energy Efficiency', 'Norm Power Efficiency']):
        ax.set_title(title, fontsize=font_title, pad=3)  # Make titles closer to figures
        ax.set_xticks(x)
        ax.set_xticklabels(x_labels, fontsize=font_tick)
        ax.tick_params(axis='y', labelsize=font_tick)
        # Make x-axis tick marks smaller on all subplots
        ax.tick_params(axis='x', length=2, width=0.3)
        ax.minorticks_off()
        ax.grid(True, linestyle='--', alpha=0.7, linewidth=0.3)
        
        # Format y-axis ticks to show integers only
        ax.yaxis.set_major_formatter(plt.FuncFormatter(lambda val, pos: f'{val:.0f}x'))
    
    # Move bottom subplot x-tick labels closer
    ax3.tick_params(axis='x', pad=1)

    # Get handles and labels from first subplot for legend
    handles, labels = ax1.get_legend_handles_labels()
    
    # Custom sorting for legend: Mugi first, then by array size, then Carat, then VA
    def legend_sort_key(item):
        label = item[1]  # label is second element in (handle, label) tuple
        
        # Extract components
        if 'Mugi' in label:
            arch_priority = 0
        elif 'Carat' in label:
            arch_priority = 1  
        elif 'VA' in label:
            arch_priority = 2
        elif 'Tensor' in label:
            arch_priority = 3
        else:
            arch_priority = 4
            
        # Extract array size for secondary sort
        import re
        size_match = re.search(r'\((\d+)\)', label)
        array_size = int(size_match.group(1)) if size_match else 0
        
        # Subarch priority (PWL, Taylor, then base)
        if 'PWL' in label:
            subarch_priority = 0
        elif 'Taylor' in label:
            subarch_priority = 1
        else:
            subarch_priority = 2
        
        # Function type (SM vs SiLU) - but for VA PWL, we want SM then SiLU
        if 'VA' in label and 'PWL' in label:
            func_priority = 0 if 'SM' in label else 1
        else:
            func_priority = 0 if 'SM' in label else 1
        
        return (arch_priority, array_size, subarch_priority, func_priority)
    
    # Sort handles and labels together
    sorted_items = sorted(zip(handles, labels), key=legend_sort_key)
    sorted_handles, sorted_labels = zip(*sorted_items)
    
    # Custom reordering to move SM VA Taylor to top of last column
    # With ncol=4, we want to find SM VA Taylor and move it to position that puts it at top of column 4
    reordered_handles = list(sorted_handles)
    reordered_labels = list(sorted_labels)
    
    # Find SM VA Taylor entry
    sm_va_taylor_idx = None
    for i, label in enumerate(reordered_labels):
        if 'SM' in label and 'VA' in label and 'Taylor' in label:
            sm_va_taylor_idx = i
            break
    
    if sm_va_taylor_idx is not None:
        # Remove SM VA Taylor from its current position
        sm_va_taylor_handle = reordered_handles.pop(sm_va_taylor_idx)
        sm_va_taylor_label = reordered_labels.pop(sm_va_taylor_idx)
        
        # Calculate position for top of last column (column 4)
        # With ncol=4, positions 0,4,8,12... are tops of columns 1,2,3,4
        total_items = len(reordered_labels) + 1  # +1 because we removed one item
        rows_needed = (total_items + 3) // 4  # Ceiling division
        target_position = 3 * rows_needed  # Top of column 4 (0-indexed)
        
        # Make sure target position doesn't exceed list length
        target_position = min(target_position, len(reordered_labels))
        
        # Insert SM VA Taylor at the target position
        reordered_handles.insert(target_position, sm_va_taylor_handle)
        reordered_labels.insert(target_position, sm_va_taylor_label)
    
    final_handles = tuple(reordered_handles)
    final_labels = tuple(reordered_labels)

    # Create legend
    fig.legend(final_handles, final_labels, ncol=4, fontsize=6, 
              loc='upper center', bbox_to_anchor=(0.55, 1.31), 
              frameon=True, columnspacing=.5, handlelength=.75, handletextpad=0.3, handleheight=0.5)

    plt.subplots_adjust(hspace=0.4)  # Increase spacing between subplots
    plt.tight_layout(pad=0.1)
    plt.savefig(output_path + 'nonlinear_breakdown.png', dpi=1200, bbox_inches='tight')
    plt.savefig(output_path + 'nonlinear_breakdown.pdf', dpi=1200, bbox_inches='tight')
    plt.show()


from archx.metric import aggregate_event_metric, aggregate_tag_metric, query_module_metric
from zoo.llm.results.query.utils import query_area, query_tag_power, load_yaml, query_dynamic_energy
import matplotlib.pyplot as plt
import pandas as pd
import os

def breakdown(df, array_colors, labels, output_path, value_unit, legend=True):
    fig_width_pt = 240  # ACM single-column width in points
    fig_width = fig_width_pt / 72  # Convert to inches
    fig_height = fig_width * 0.65  # Given height

    n_columns = len(df)
    n_unique_columns = (df['arch'].nunique())
    n_rows = n_columns // n_unique_columns

    n_rows = 2
    n_unique_columns = 5

    fig, axes = plt.subplots(nrows=n_rows, ncols=n_unique_columns, figsize=(fig_width, fig_height),
                             gridspec_kw={"wspace": -0.3, "hspace": 0.2, 'left': -0.65, 'right': 0.65})
    
    for index, row in df.iterrows():
        ax_row = index % n_rows
        ax_column = index // n_rows
        
        arch = row['arch']
        subarch = row['subarch']
        arch_dim = row['arch_dim']
        total = row['total']
        values = row.drop(labels=['arch', 'subarch', 'arch_dim', 'total']).to_list()

        arch = arch.capitalize()
        arch = 'SA' if arch == 'Systolic' else 'SD' if arch == 'Simd' else arch
        subarch = '' if subarch == 'mac' else '-F' if subarch == 'figna' else '-L' if subarch == 'lut' else ''
        arch_dim = '(' + arch_dim.split('x')[0] + ')'

        ax = axes[ax_row, ax_column]

        ax.pie(
            values,
            labels=None,
            autopct=lambda p: f'{p:.0f}%' if p > 7 else '',
            startangle=90,
            colors=array_colors,
            textprops={'fontsize': 5}
        )
        ax.set_title(f"{arch}{subarch} {arch_dim}\n({total:.2f} {value_unit})", fontsize=5, pad=-2)
        ax.axis('off')

    if (legend == True):
        plt.legend(
            labels, loc='lower center', ncol=7, fontsize=5, bbox_to_anchor=(-1.45, -0.2)
        )
    plt.savefig(output_path, dpi=1200, bbox_inches='tight')
    plt.close()

def figure(input_path: str, output_path: str):

    area_df = pd.read_csv(input_path + 'array_area_breakdown.csv')
    power_df = pd.read_csv(input_path + 'array_power_breakdown.csv')

    labels = ['Fifo', 'PE', 'TC', 'Value Reuse', 'Acc', 'Nonlinear', 'Vector']
    array_colors = ['#D5AAFF', '#A8D5BA', "#EBE596", '#85E3FF', '#f09494', '#FFC3A0', "#a7f672"]

    breakdown(df=area_df, array_colors=array_colors, labels=labels, output_path=output_path + 'area_breakdown.pdf', value_unit='mm²', legend=False)
    breakdown(df=power_df, array_colors=array_colors, labels=labels, output_path=output_path + 'power_breakdown.pdf', value_unit='mW')

def query(input_path, output_path):

    vlp_list = ['mugi', 'carat']
    vlp_arch_dim_list = ['128x8', '256x8']

    mugi_subarch_list = ['vlp', 'lut']

    baseline_list = ['systolic', 'simd']
    baseline_arch_dim_list = ['8x8', '16x16']
    baseline_subarch_list = ['figna']

    network = 'single_node'
    model = 'llama_2_7b'
    max_seq_len = 'max_seq_len_4096'
    batch_size = 'batch_size_8'
    
    tag_list = ['fifo', 'pe', 'tc', 'value_reuse', 'accumulator', 'nonlinear', 'vector']
    df_labels = ['arch', 'subarch', 'arch_dim', 'total'] + tag_list
    area_df = pd.DataFrame(columns=df_labels)
    power_df = pd.DataFrame(columns=df_labels)

    for arch in vlp_list + baseline_list:
        for subarch in (baseline_subarch_list if arch in baseline_list else mugi_subarch_list if arch in ['mugi'] else ['']):
            for arch_dim in (vlp_arch_dim_list if arch in vlp_list else baseline_arch_dim_list if arch in baseline_list else ['8x16x16'] if arch == 'tensor' else ['']):
            
                termination_path = 'full_termination' if arch == 'mugi' else ''
                run_path = os.path.normpath(f'{input_path}{arch}/{network}/{subarch}/{arch_dim}/{model}/{max_seq_len}/{batch_size}/{termination_path}/')
                yaml_dict = load_yaml(run_path)

                event_graph = yaml_dict['event_graph']
                metric_dict = yaml_dict['metric_dict']

                if arch == 'mugi' and subarch == 'vlp' and arch_dim == '128x8':
                    fifo_area = query_area(event_graph=event_graph, metric_dict=metric_dict, workload=model, module='pe_fifo') 

                area_row = {
                    'arch': arch,
                    'subarch': subarch,
                    'arch_dim': arch_dim,
                }

                power_row = {
                    'arch': arch,
                    'subarch': subarch,
                    'arch_dim': arch_dim,
                }

                total_area = 0
                total_power = 0
                for tag in tag_list:
                    try:
                        tag_area = query_area(tag=tag, event_graph=event_graph, metric_dict=metric_dict, workload=model)
                        tag_power = query_tag_power(tag=tag, event_graph=event_graph, metric_dict=metric_dict, workload=model, event=model)
                    except:
                        tag_area = 0
                        tag_power = 0
                    total_area += tag_area
                    total_power += tag_power
                    area_row[tag] = tag_area
                    power_row[tag] = tag_power

                # tag_array_area = query_area(tag='array', event_graph=event_graph, metric_dict=metric_dict, workload=model)
                # tag_array_power = query_tag_power(tag='array', event_graph=event_graph, metric_dict=metric_dict, workload=model, event=model)

                area_row['total'] = total_area
                power_row['total'] = total_power
                # area_row['array'] = tag_array_area
                # power_row['array'] = tag_array_power

                area_df = pd.concat([area_df, pd.DataFrame([area_row])], ignore_index=True) if not area_df.empty else pd.DataFrame([area_row])
                power_df = pd.concat([power_df, pd.DataFrame([power_row])], ignore_index=True) if not power_df.empty else pd.DataFrame([power_row])
    
    area_df.to_csv(output_path + 'array_area_breakdown.csv', index=False)
    power_df.to_csv(output_path + 'array_power_breakdown.csv', index=False)


from zoo.llm.results.query.utils import query_throughput_energy_metrics, load_yaml, compute_throughput, geomean
import pandas as pd
from matplotlib import pyplot as plt
import numpy as np
from collections import OrderedDict
import os

def query(input_path, output_path):
    
    network_list = ['single_node']

    mugi_subarch_list = ['vlp']

    vlp_list = ['mugi', 'carat']
    vlp_arch_dim_list = ['64x8', '256x8']

    baseline_list = ['systolic', 'simd']
    baseline_arch_dim_list = ['8x8', '16x16']
    baseline_subarch_list = ['mac', 'figna']

    throughput_module = OrderedDict({
        'mugi': OrderedDict({'gemm': 'and_gate', 'nonlinear': 'and_gate'}),
        'carat': OrderedDict({'gemm': 'and_gate', 'nonlinear': 'register_vector'}),
        'systolic': OrderedDict({'gemm': 'multiplier', 'nonlinear': 'register_vector'}),
        'simd': OrderedDict({'gemm': 'multiplier', 'nonlinear': 'register_vector'}),
        'tensor': OrderedDict({'gemm': 'multiplier', 'nonlinear': 'register_vector'}),
    })

    batch_size = 'batch_size_8'

    model_list = ['llama_2_70b_GQA']
    max_seq_len_list = ['max_seq_len_128', 'max_seq_len_256', 'max_seq_len_512', 'max_seq_len_1024', 'max_seq_len_2048', 'max_seq_len_4096']
    batch_size_list = ['batch_size_1', 'batch_size_2', 'batch_size_4', 'batch_size_8', 'batch_size_16', 'batch_size_32']
    gqa_size_list = ['kv_heads_64', 'kv_heads_32', 'kv_heads_16', 'kv_heads_8']

    workload = 'llama_2'

    attention_size_breakdown_df = pd.DataFrame()

    for network in network_list:
        for arch in (vlp_list + baseline_list):
            for subarch in (baseline_subarch_list if arch in baseline_list else mugi_subarch_list if arch in ['mugi'] else ['']):
                subarch_label = 'none' if subarch == '' else subarch
                for arch_dim in (vlp_arch_dim_list if arch in vlp_list else baseline_arch_dim_list if arch in baseline_list else ['8x16x16'] if arch in ['tensor'] else ['']):
                    for max_seq_len in max_seq_len_list:
                        for gqa_size in gqa_size_list:
                            for model in model_list:

                                gemm_module = throughput_module[arch]['gemm']

                                termination_path = 'full_termination' if arch == 'mugi' else ''
                                run_path = os.path.normpath(f'{input_path}{arch}/{network}/{subarch}/{arch_dim}/{model}/{max_seq_len}/{batch_size}/{gqa_size}/{termination_path}/')
                                yaml_dict = load_yaml(run_path)

                                event_graph = yaml_dict['event_graph']
                                metric_dict = yaml_dict['metric_dict']

                                gemm_performance_metrics_dict = query_throughput_energy_metrics(event_graph=event_graph, metric_dict=metric_dict, workload=workload, event='attention', module=gemm_module)

                                performance_metrics_dict = OrderedDict({
                                    'flops': gemm_performance_metrics_dict['flops'],
                                    'execution_time': gemm_performance_metrics_dict['execution_time']
                                })

                                throughput_eff_dict = compute_throughput(performance_metrics_dict=performance_metrics_dict)
                                throughput_eff_dict['energy_per_token'] = gemm_performance_metrics_dict['energy'] / int(max_seq_len.split('_')[-1])

                            attention_size = 'attn_size_' + str(64 // int(gqa_size.split('_')[-1]))

                            attention_size_breakdown_dict = OrderedDict({
                                'arch': arch,
                                'subarch': subarch_label,
                                'arch_dim': arch_dim,
                                'max_seq_len': max_seq_len,
                                'attn_size': attention_size,
                                'throughput': throughput_eff_dict['throughput'],
                                'energy_per_token': throughput_eff_dict['energy_per_token'],
                                # 'flops': throughput_eff_geomean_dict['flops'],
                                # 'execution_time': throughput_eff_geomean_dict['execution_time'],
                            })

                            attention_size_breakdown_df = pd.concat([attention_size_breakdown_df, pd.DataFrame(attention_size_breakdown_dict, index=[0])])

    attention_size_breakdown_df.to_csv(output_path + 'attention_size_breakdown.csv', index=False)

    baseline_df = attention_size_breakdown_df[
        (attention_size_breakdown_df['arch'] == 'systolic') &
        (attention_size_breakdown_df['subarch'] == 'mac') &
        (attention_size_breakdown_df['arch_dim'] == '8x8') &
        (attention_size_breakdown_df['attn_size'] == 'attn_size_1')
    ]

    merge_list = ['max_seq_len']
    numeric_columns = baseline_df.select_dtypes(include=['number']).columns
    columns_to_merge = merge_list + list(numeric_columns)

    merged_df = attention_size_breakdown_df.merge(
        baseline_df[columns_to_merge],
        on=merge_list,
        suffixes=('', '_baseline')
    )

    merged_df['throughput'] = merged_df['throughput'] / merged_df['throughput_baseline']
    merged_df['energy_per_token'] = merged_df['energy_per_token'] / merged_df['energy_per_token_baseline']
    normalized_df = merged_df.drop(
        columns=['throughput_baseline', 'energy_per_token_baseline']
    )

    normalized_df.to_csv(output_path + 'attention_size_norm.csv', index=False)

def figure(input_path: str, output_path: str):
    df = pd.read_csv(input_path + 'attention_size_norm.csv')
    
    mugi_df = df[df['arch'] == 'mugi']
    other_df = df[df['arch'] != 'mugi']

    df = pd.concat([other_df, mugi_df])

    fig_width_pt = 500              
    fig_width = fig_width_pt / 72   
    fig_height = fig_width / 4.75  # Doubled height for two subplots
    fig, (ax1, ax2) = plt.subplots(2, 1, figsize=(fig_width, fig_height), sharex=True, 
                                   gridspec_kw={'hspace': 0.6, 'height_ratios': [0.7, 1]})

    color_map = {
        'Mugi': "#7BCCEC",
        'Carat': "#E4DD86",
        'SA': "#63C78A",
        'SA-F': "#C27CF1",
        'SD': "#F3AB58",
        'SD-F': "#FF9B9B",
    }
    
    marker_map = {
        '64x8': 'v',
        '256x8': 's',
        '8x8': 'v',
        '16x16': 's'
    }
    
    attn_sizes = [1, 2, 4, 8]
    seq_lens = sorted([int(seq.split('_')[-1]) for seq in df['max_seq_len'].unique()])
    segment_spacing = len(attn_sizes) + 2
    
    x_positions = []
    all_labels = []
    
    # Plot both throughput and energy_per_token
    seq_len_positions = []  # Store positions for sequence length labels
    seq_len_labels = []     # Store sequence length labels
    
    for ax, metric, title, y_lim, y_ticks in [(ax1, 'throughput', 'Normalized Throughput', (0, 25), [5, 15, 25]), 
                                               (ax2, 'energy_per_token', 'Normalized Energy per Token', (0.25, 1.3), [0.25, 0.75, 1.25])]:
        
        for j, seq_len in enumerate(df['max_seq_len'].unique()):
            seq_len_value = int(seq_len.split('_')[-1])
            seq_len_df = df[df['max_seq_len'] == seq_len]
            start_pos = j * segment_spacing
            x_positions_segment = np.arange(len(attn_sizes)) + start_pos
            if ax == ax1:  # Only populate x_positions and seq_len info once
                x_positions.extend(x_positions_segment)
                mid_position = np.mean(x_positions_segment)
                seq_len_positions.append(mid_position)
                seq_len_labels.append(f'{seq_len_value}')
            
            for arch in seq_len_df['arch'].unique():
                arch_df = seq_len_df[seq_len_df['arch'] == arch]
                
                for subarch in arch_df['subarch'].unique():
                    subarch_df = arch_df[arch_df['subarch'] == subarch]
                    
                    # Sort arch_dim so squares ('256x8', '16x16') are plotted first, then triangles ('128x8', '8x8')
                    arch_dims = sorted(subarch_df['arch_dim'].unique(), 
                                     key=lambda x: 0 if marker_map.get(x, 'o') == 's' else 1)
                    
                    for arch_dim in arch_dims:
                        arch_dim_df = subarch_df[subarch_df['arch_dim'] == arch_dim]
                        #arch_label = 'Mugi/Carat' if arch == 'mugi' else 'SA/SA-F' if arch == 'systolic' else 'SD/SD-F'
                        arch_label = 'Mugi' if arch == 'mugi' else 'Carat' if arch == 'carat' else 'SA' if arch == 'systolic' else 'SD'
                        
                        # Add subarch to label
                        if subarch == 'figna':
                            arch_label += '-F'
                        
                        #if arch_label == 'SA' and subarch
                        label = f"{arch_label} ({arch_dim.split('x')[0]})" if j == 0 and ax == ax1 else None
                        if j == 0 and ax == ax1:
                            all_labels.append(label)
                        color = color_map.get(arch_label, 'black')
                        marker = marker_map.get(arch_dim, 'o')
                        y_data = []

                        for attns in attn_sizes:
                            attn_size = f'attn_size_{attns}'
                            value = arch_dim_df[arch_dim_df['attn_size'] == attn_size][metric].values
                            y_data.append(value[0] if len(value) > 0 else np.nan)
                        
                        ax.plot(
                            x_positions_segment,
                            y_data,
                            label=label,
                            color=color,
                            marker=marker,
                            linestyle='-',
                            markersize=2.5,
                            linewidth=0.6,
                            alpha=0.9,
                            markeredgecolor='black',
                            markeredgewidth=0.1,
                            markerfacecolor=color
                        )
        
        # Format axis
        if metric == 'throughput':
            ax.yaxis.set_major_formatter(plt.FuncFormatter(lambda val, pos: f'{int(val)}x'))
        else:
            ax.yaxis.set_major_formatter(plt.FuncFormatter(lambda val, pos: f'{val:.2f}x'))
        
        ax.grid(axis='y', linestyle='--', color='grey', alpha=0.5, linewidth=0.5)
        if(metric == 'throughput'):
            ax.set_title(title, fontsize=7, pad = 13)
        else:
            ax.set_title(title, fontsize=7)
        ax.set_ylim(y_lim)
        ax.tick_params(axis='y', labelsize=7)
        ax.set_yticks(y_ticks)
        
        # Add vertical lines at sequence length positions
        for pos in seq_len_positions:
            ax.axvline(x=pos, color='gray', linestyle=':', alpha=0.3, linewidth=0.5)
    
    # Create secondary x-axis on top subplot for sequence length labels
    ax1_top = ax1.twiny()
    ax1_top.set_xlim(ax1.get_xlim())
    ax1_top.set_xticks(seq_len_positions)
    ax1_top.set_xticklabels(seq_len_labels, fontsize=5.5)
    ax1_top.tick_params(axis='x', length=3, pad=1)  # Show tick marks on top
    
    # Create secondary x-axis on bottom subplot for sequence length tick marks only (no labels)
    ax2_top = ax2.twiny()
    ax2_top.set_xlim(ax2.get_xlim())
    ax2_top.set_xticks(seq_len_positions)
    ax2_top.set_xticklabels([])  # No labels, just tick marks
    ax2_top.tick_params(axis='x', length=3, top=True, labeltop=False)  # Show tick marks but no labels
    
    # Only set x-axis labels on bottom plot
    ax2.set_xticks(x_positions)
    xtick_labels = [f'{attns}' for attns in attn_sizes] * len(seq_lens)
    ax2.set_xticklabels(xtick_labels, fontsize=5.5)
    
    # Get handles and labels from the plot, then sort them together
    handles, labels = ax1.get_legend_handles_labels()
    
    # Create pairs of (handle, label) and sort them
    handle_label_pairs = [(handle, label) for handle, label in zip(handles, labels) if label is not None]
    
    # Sort pairs by architecture and then by array size (smaller first)
    def sort_pairs(pair):
        label = pair[1]
        
        # Extract array size from label (e.g., "Mugi (128)" -> 128)
        import re
        size_match = re.search(r'\((\d+)\)', label) if label else None
        array_size = int(size_match.group(1)) if size_match else 999
        
        if label and 'Mugi' in label:
            return (0, array_size, label)  # Mugi first, then by size
        elif label and 'Carat' in label:
            return (1, array_size, label)  # Carat second, then by size
        elif label and 'SA-F' in label:
            return (3, array_size, label)  # SA-F fourth, then by size
        elif label and 'SA' in label:
            return (2, array_size, label)  # SA third, then by size
        elif label and 'SD-F' in label:
            return (5, array_size, label)  # SD-F sixth, then by size
        elif label and 'SD' in label:
            return (4, array_size, label)  # SD fifth, then by size
        else:
            return (6, array_size, label)  # Others last
    
    sorted_pairs = sorted(handle_label_pairs, key=sort_pairs)
    sorted_handles = [pair[0] for pair in sorted_pairs]
    sorted_labels = [pair[1] for pair in sorted_pairs]
    
    # fig.legend(
    #     sorted_handles,
    #     sorted_labels,
    #     frameon=True,
    #     loc='upper center',
    #     fontsize=7,
    #     ncol=6,
    #     bbox_to_anchor=(0.5115, 1.35),
    #     columnspacing=1.15
    # )
    
    plt.tight_layout(pad=0.2)
    plt.savefig(output_path + 'attention_size_breakdown.png', dpi=1200, bbox_inches='tight')
    plt.savefig(output_path + 'attention_size_breakdown.pdf', dpi=1200, bbox_inches='tight')
    plt.show()